unit carterll;

interface

  uses ftpermin; { this unit defines constants }

{ Carter coordinate to Latitude/Longitude conversion routines

  adapted by: Brandon C. Nuttall

              Kentucky Geological Survey
              228 Mining & Mineral Resources Building
              University of Kentucky
              Lexington, KY  40506-0107

  Language:   Borland International TURBO Pascal v3.01a
                (This source code no longer compatible with v3.01a)
              Borland International TURBO Pascal v4.0
              Borland International TURBO Pascal v5.0 (3/3/89 BCN)
  Computer:   DEC Rainbow PC 100B (512K)
  Operating system: CP/M-86/80 v2.11
                    MS-DOS v3.1 (for TURBO Pascal v5.0)

  Adapted from:
    Fortran version implemented by Kentucky Geological Survey
    in public domain

**************************************************************************
*                                                                        *
*                            N O T I C E                                 *
*                                                                        *
*     This software is placed in the public domain by the Kentucky       *
*     Geological Survey and is supplied "as is".                         *
*                                                                        *
*     The Kentucky Geological Survey makes no warranty either express    *
*     or implied as to the accuracy or validity of these routines.  The  *
*     user assumes all liability for any losses or damages that may be   *
*     incurred through any use or application of this software.          *
*                                                                        *
**************************************************************************

  Note:  For complete documentation see the file CARTERLL.DOC supplied
         on the distribution diskette.

  Revision History:
    14-Mar-1986 -- 1) Bug:  In FeetFirst function, dash between
                   section and letter not detected caused unexpected
                   dash error and inability to determine format.
                      Fix:  Let this dash be a valid indicator that
                   the first part of the input string is not the footages.

                   2) KGS id, program title, and program version messages
                   added.

                   3) Sample program removed and implemented as compile-ready
                   source code.

    01-Jun-1988 -- 1) Converted to TURBO Pascal Unit and compiled under
                   MS-DOS

    19-May-1989 -- Add routine to convert latitude/longitude to Carter
                   coordinates

                   1) replace type "anystr" with TURBO Pascal type string
                   2) define constants for latitude and longitude base
                      and reference lines

       Jun-1989 -- Complete debugging and add _InKentucky check of
                   valid Carter coordinates

    23-Jun-1989 -- Bug:  if CarterToLatLon() bailed out before call
                   to _CarterToLatLon() its return value was unknown
                   and likely to be wrong.  Fix: initialize
                   CarterToLatLon() to false;  the only way it can
                   be true now is if _CarterToLatLon() succeeds.

     8-Jul-1994 -- Treat tab characters in input like blanks

***********************************************************************
                         K N O W N   B U G S
***********************************************************************

     The Carter coordinate system was originally defined using a base
latitude of 36deg 30min north.  Any areas south of that base latitude
do not have a defined Carter coordinate.  Some minor areas of extreme
western Kentucky along the Kentucky/Tennessee border in the Jackson
Purchase are south of the base latitude.  At this time, there is no
convention for a Carter coordinate 5-minute section letter below
"A"; a Carter coordinate cannot be input.  Additionally, latitudes
south of 36.5 degrees are not recognized as being in Kentucky.  Be
forewarned.

}

{ ************************************************************************
  global definitions for Carter coordinate to latitude/longitude conversion }

const
  ccllsource = 'Kentucky Geological Survey, University of Kentucky';
  ccllname   =
    'Carter Coordinate to Latitude/Longitude Conversion, v4.1, 8-Jun-1994';

type
  Sections    = 1..25;
  CarterLtrs  = (asym, bsym, csym, dsym, esym, fsym, gsym, hsym, isym, jsym,
                 ksym, lsym, msym, nsym, osym, psym, qsym, rsym, ssym, tsym,
                 usym, vsym, wsym, xsym, ysym, zsym, aasym, bbsym, ccsym,
                 ddsym, eesym, ffsym, ggsym);
  CLArray     = array[CarterLtrs] of string[2];
  Numbers     = 0..91;
  boolptr     = ^boolean;
  charset     = set of char;

var
  CarterFormKnown          : boolptr;

const { typed constants here }

{ NOTICE: these are TURBO Pascal "typed constants" and as such may
  not be portable to other Pascal compilers. }

  CarterRep : CLArray = ('A','B','C','D','E','F','G','H','I','J','K','L',
                         'M','N','O','P','Q','R','S','T','U','V','W','X',
                         'Y','Z','AA','BB','CC','DD','EE','FF','GG');

  debug : boolean = false; { tell the user some things if they want
                             to know }

  expand : boolean = false; { set up the output Carter coordinate format
                              flag to the Survey default }

{ Routines starting with underscore character ("_") are strongly typed
  workhorse routines.  Applications programmers that want the easy way
  out should use the primary routines }

function _InKentucky(Identifier : string;
                     FiveMinSecLetter : CarterLtrs;
                     FiveMinSecNumber : Numbers) : boolean;

function _CCToLatLon(Identifier       : string; { something to identify data }
                     FeetFromNSLine   : integer;
                     NSLineLetter     : char;   { single letter N or S }
                     FeetFromEWLine   : integer;
                     EWLineLetter     : char;   { single letter E or W }
                     OneMinSecNumber  : Sections;
                     FiveMinSecLetter : CarterLtrs; { user defined type }
                     FiveMinSecNumber : Numbers;
                     var Latitude, Longitude : real) : boolean;

function _LatLonToCC(Identifier       : string; { something to identify data }
                     var FeetFromNSLine   : integer;
                     var NSLineLetter     : char;   { single letter N or S }
                     var FeetFromEWLine   : integer;
                     var EWLineLetter     : char;   { single letter E or W }
                     var OneMinSecNumber  : Sections;
                     var FiveMinSecLetter : CarterLtrs; { user defined type }
                     var FiveMinSecNumber : Numbers;
                     Latitude, Longitude  : real) : boolean;

{ these routines are the primary routines for doing the conversions }

function InKentucky(Identifier, SecLetter : string;
                    SecNumber : integer) : boolean;

function CarterToLatLon(Identifier, CarterCoord : string;
                        var Latitude, Longitude : real) : boolean;

function LatLonToCarter(Identifier : string; var CarterCoord : string;
                        Latitude, Longitude : real;
                        expand : boolean) : boolean;

implementation

{ these constants are local to the unit }
const

  cclatorg   = 2190; { latitude in minutes of origin of Carter coordinate
                       grid (lat 36deg 30min N) }
  cclatref   = 2160; { reference latitude in minutes for feet per minute
                       of latitude lookup tables }
  cclonorg   = 5375; { longitude in minutes of origin of Carter coordinate
                       grid (lon 89deg 35min W)}
  cclonref   = 4920; { reference longitude in minutes (lon 82deg W)}

{ function CvtToLtrSym -- LOCAL TO CARTERLL UNIT

  Purpose: to convert the ASCII upper case letter(s) in the input token
      into a form used internally for calculation (something the ordinal
      value can be determined for)

  Usage: MAY NOT BE CALLED BY THE USER AS IMPLEMENTED

  Errors detected:
      1) invalid Carter coordinate 5-minute quad letter

  Returns: boolean value TRUE iff input is a valid quad letter

  Note:  If the user decides to make this function available outside the
      unit by declaring it in the interface section, code should be
      added to standardize the input letter to upper case and strip
      leading and trailing blanks.

}

function CvtToLtrSym(Identifier : string;
                     token : string; var Sym : CarterLtrs) : boolean;
  var
    Ok : boolean;
    temp : CarterLtrs;
  begin
    Ok := false;
    Sym := asym;
    for Temp := asym to ggsym do
      if token=CarterRep[temp] then begin
        Ok  := true;
        Sym := temp
      end;
    if Ok
      then CvtToLtrSym := true
      else begin
        writeln('Invalid Carter coordinate 5-minute letter "',token,
                '" at id ',Identifier);
        CvtToLtrSym := false
      end;
  end; { CvtToLtrSym }


{ *********************************************************************
  function _InKentucky

  Purpose: Do gross check of valid Carter coordinate 5-minute section
      letter and number ranges.

  Usage:

      boolean_variable := _InKentucky(identifier_string,
                                      letter_value,
                                      number_value);

      identifier_string -- any string containing information useful for
          identifying which item is being processed

      letter_value -- a value of type CarterLtrs (user defined enumerated
          type) representing the Carter coordinate 5-minute section
          letter

      number_value -- a value of type Numbers (i.e., 0 to 91) representing
          the Carter coordinate 5-minute section number

  Errors detected:

      Prints message when input values are not in Kentucky.

      Note:
          1) Although letter GG is defined as a valid Carter coordinate
             5-minute section letter, all GG locations are north of
             Kentucky and will return an error.
          2) Some areas along the Tennessee/Kentucky border in the
             Jackson Purchase of western Kentucky do not have a valid
             Carter coordinate.  For Kentucky, no coordinate was defined
             for any areas south of 36deg 30min north latitude

  Returns: boolean value TRUE iff in Kentucky (as noted above)
}
function _InKentucky;
var
  ok : boolean;
begin
  ok := false;
  case FiveMinSecLetter of
     asym : if ((FiveMinSecNumber >=  0) and (FiveMinSecNumber <= 18)) or
               ((FiveMinSecNumber >= 69) and (FiveMinSecNumber <= 70))
              then ok := true;
     bsym : if (FiveMinSecNumber >=  0) and (FiveMinSecNumber <= 75)
              then ok := true;
     csym : if (FiveMinSecNumber >=  4) and (FiveMinSecNumber <= 77)
              then ok := true;
     dsym,
     esym : if (FiveMinSecNumber >=  4) and (FiveMinSecNumber <= 80)
              then ok := true;
     fsym : if (FiveMinSecNumber >=  4) and (FiveMinSecNumber <= 81)
              then ok := true;
     gsym : if (FiveMinSecNumber >=  4) and (FiveMinSecNumber <= 82)
              then ok := true;
     hsym : if (FiveMinSecNumber >=  4) and (FiveMinSecNumber <= 84)
              then ok := true;
     isym : if (FiveMinSecNumber >=  5) and (FiveMinSecNumber <= 86)
              then ok := true;
     jsym : if (FiveMinSecNumber >= 12) and (FiveMinSecNumber <= 88)
              then ok := true;
     ksym : if (FiveMinSecNumber >= 12) and (FiveMinSecNumber <= 89)
              then ok := true;
     lsym : if (FiveMinSecNumber >= 13) and (FiveMinSecNumber <= 90)
              then ok := true;
     msym : if (FiveMinSecNumber >= 17) and (FiveMinSecNumber <= 91)
              then ok := true;
     nsym : if (FiveMinSecNumber >= 17) and (FiveMinSecNumber <= 89)
              then ok := true;
     osym,
     psym : if (FiveMinSecNumber >= 17) and (FiveMinSecNumber <= 87)
              then ok := true;
     qsym : if (FiveMinSecNumber >= 17) and (FiveMinSecNumber <= 86)
              then ok := true;
     rsym : if ((FiveMinSecNumber >= 19) and (FiveMinSecNumber <= 26)) or
               ((FiveMinSecNumber >= 30) and (FiveMinSecNumber <= 85))
              then ok := true;
     ssym : if (FiveMinSecNumber >= 36) and (FiveMinSecNumber <= 84)
              then ok := true;
     tsym : if ((FiveMinSecNumber >= 37) and (FiveMinSecNumber <= 39)) or
               ((FiveMinSecNumber >= 44) and (FiveMinSecNumber <= 84))
              then ok := true;
     usym : if ((FiveMinSecNumber >= 38) and (FiveMinSecNumber <= 39)) or
               ((FiveMinSecNumber >= 44) and (FiveMinSecNumber <= 84))
              then ok := true;
     vsym : if (FiveMinSecNumber >= 44) and (FiveMinSecNumber <= 84)
              then ok := true;
     wsym,
     xsym : if (FiveMinSecNumber >= 47) and (FiveMinSecNumber <= 84)
              then ok := true;
     ysym : if (FiveMinSecNumber >= 49) and (FiveMinSecNumber <= 83)
              then ok := true;
     zsym : if (FiveMinSecNumber >= 49) and (FiveMinSecNumber <= 81)
              then ok := true;
    aasym : if (FiveMinSecNumber >= 49) and (FiveMinSecNumber <= 80)
              then ok := true;
    bbsym : if ((FiveMinSecNumber >= 54) and (FiveMinSecNumber <= 68)) or
               ((FiveMinSecNumber >= 79) and (FiveMinSecNumber <= 80))
              then ok := true;
    ccsym,
    ddsym : if (FiveMinSecNumber >= 56) and (FiveMinSecNumber <= 64)
              then ok := true;
    eesym : if (FiveMinSecNumber >= 56) and (FiveMinSecNumber <= 63)
              then ok := true;
    ffsym : if (FiveMinSecNumber >= 56) and (FiveMinSecNumber <= 61)
              then ok := true;
    { note that no gg coordinates are really in Kentucky }
    ggsym : ok := false;
  end; { case }
  if not ok then
    writeln('5-minute section, "',CarterRep[FiveMinSecLetter],
            '-',FiveMinSecNumber,'", is not in Kentucky at id ',
            Identifier);
  _InKentucky := ok;
end; { _InKentucky }

{ ************************************************************************
  function _CCToLatLon

  Purpose: Main Carter coordinate to latitude/longitude conversion
      routine.

  Usage:

      boolean_variable := _CCToLatLon(...);

      May be called by the user directly.  However, input to this routine
      must be as specified and of the given types.

  Errors detected:
      1) The magnitude of the footage values are checked
      2) Any error during initialization causes program termination
      3) Checks letter and number for "in Kentucky" range

  Returns: boolean value TRUE if and only if Carter coordinate to latitude
      longitude conversion succeeds

  Note:
      If your data is of the types required by this routine:
        1) _CCToLatLon is much the fastest and most efficient way to do the
           conversion.
        2) CarterToLatLon is unnecessary.
}

function _CCtoLatLon;
type
  OffsetArray = array[Sections] of integer;
var
  FeetFromSouthLine, FeetFromEastLine,
  FtPerMinLat, FtPerMinLon, S         : real;
  eedge, sedge, miscedge, MN36, MI82  : integer;
const
  LonOffset : OffsetArray = (0,1,2,3,4,4,3,2,1,0,0,1,2,3,4,4,3,2,1,0,
                             0,1,2,3,4);
  LatOffset : OffsetArray = (4,4,4,4,4,3,3,3,3,3,2,2,2,2,2,1,1,1,1,1,
                             0,0,0,0,0);
begin { _CCToLatLon }
  Latitude  := 0.0; { always set to 0 so there is a known value returned }
  Longitude := 0.0;
  if not _InKentucky(Identifier,FiveMinSecLetter,FiveMinSecNumber) then begin
    _CCToLatLon := false;
    exit;
  end;
  { get latitude in minutes from south edge of 1-minute quad from south edge
    of state (Carter coordinate origin) }
  miscedge := 5 * ord(FiveMinSecLetter) + LatOffset[OneMinSecNumber];
  { get latitude in minutes from south edge of 1-minute quad from equator }
  sedge := cclatorg + miscedge;
  { find feet per minute of latitude on ground at this latitude }
  MN36 := sedge - cclatref;
  FtPerMinLat := LatFtPerMin[MN36];
  if FeetFromNSLine > FtPerMinLat then begin
    writeln('Footage from north/south line is out of range at id ',
            Identifier);
    _CCToLatLon := false;
    exit
  end;
  if NSLineLetter = 'N'
    then FeetFromSouthLine := FtPerMinLat - FeetFromNSLine
    else FeetFromSouthLine := FeetFromNSLine;
  S := ((FeetFromSouthLine*1.0) / (FtPerMinLat*1.0)) * 60.0;
  Latitude := (sedge*1.0)/60.0 + S/3600.0;
  { get longitude of east edge of 5-minute quad in minutes from 82 degree
    line }
  MI82 := 5 * (90 - FiveMinSecNumber);
  { longitude in minutes of eastern edge of 1-minute quad from prime meridian }
  eedge := cclonref + MI82 + LonOffset[OneMinSecNumber];
  FtPerMinLon := LonFtPerMin[MN36];
  if FeetFromEWLine > FtPerMinLon then begin
    writeln('Footage from east/west line is out of range at id ',
            Identifier);
    _CCToLatLon := false;
    exit
  end;
  if EWLineLetter = 'W'
    then FeetFromEastLine := FtPerMinLon - FeetFromEWLine
    else FeetFromEastLine := FeetFromEWLine;
  S := ((FeetFromEastLine*1.0) / ( FtPerMinLon*1.0)) * 60.0;
  Longitude := -1.0*((eedge*1.0)/60.0 + S/3600.0);
  _CCToLatLon := true
end; { _CCToLatLon }

{ **********************************************************************
  function _LatLonToCC

  Purpose: Main routine for conversion of latitude/longitude to Carter
      coordinate.

  Usage:

      boolean_variable := _LatLonToCC(...);

      May be called by the user directly.  However, output from this
      routine is in a variety of user defined types and as such is not
      directly printable.

  Errors detected:

      1) The magnitude of the input latitude and longitude values are
         checked against their ranges in Kentucky.
      2) The calculated Carter coordinate 5-minute letter and number are
         checked for "in Kentucky"

  Returns:

      Boolean value TRUE is returned if and only if the conversion to
      Carter coordinates succeeds.

  Notes:

      If you want to format the output Carter coordinate string other
      than the manner provided by the function LatLonToCarter, use this
      routine to do all your conversions and format it yourself.
}
function _LatLonToCC;
type
  SectionArry = array[0..4,0..4] of Sections;
const
  SectionNames : SectionArry = ((25,24,23,22,21), { look up sections }
                                (16,17,18,19,20),
                                (15,14,13,12,11),
                                ( 6, 7, 8, 9,10),
                                ( 5, 4, 3, 2, 1));
var
  latmin, lonmin,        { integer minutes of latitude and longitude }
  mn36        : integer; { minutes north of latitude 36deg N }
  FtPerMinLat,           { feet on ground per minute of latitude for
                         the target latitude }
  FtPerMinLon : real;    { feet on ground per minute of longitude for
                         the target latitude }
  fraclatmin, fraclonmin : real; { fractional minutes of lat and long }

  { procedure to convert decimal degrees to integer minutes and fractional
    minutes }
  procedure ToMinutes(degrees : real; var minutes : integer;
                                      var fraction : real);
  begin
    minutes := trunc(degrees * 60.0);
    fraction := (degrees * 60.0) - minutes;
  end;

begin { _LatLonToCC }
  { initialize to return '0 0 25 A0' on invalid input }
  FeetFromNSLine   := 0;
  NSLineLetter     := ' ';
  FeetFromEWLine   := 0;
  EWLineLetter     := ' ';
  OneMinSecNumber  := 25;
  FiveMinSecLetter := asym;
  FiveMinSecNumber := 0;
  { make longitude positive if needed }
  if Longitude < 0.0 then Longitude := Longitude * -1.0;
  { do a quick check of range of latitude and longitude }
  if (Longitude > 89.58333333) or (Longitude < 81.93333333) then begin
    Writeln('Longitude ',Longitude:10:6,' is outside of Kentucky for ',
            'identifier = ',identifier);
    _LatLonToCC := false;
    exit;
  end;
  if (Latitude > 39.16666667) or (Latitude < 36.48333333) then begin
    Writeln('Latitude ',Latitude:10:6,' is outside of Kentucky for ',
            'identifier = ',identifier);
    _LatLonToCC := false;
    exit;
  end;
  { calculate Carter letter, number, and section }
  ToMinutes(Latitude,latmin,fraclatmin);
  ToMinutes(Longitude,lonmin,fraclonmin);
  FiveMinSecLetter := CarterLtrs((latmin - cclatorg) div 5);
  FiveMinSecNumber := (cclonorg - lonmin - 1) div 5;
  if not _InKentucky(Identifier,FiveMinSecLetter,FiveMinSecNumber) then begin
    _LatLonToCC := false;
    exit;
  end;
  OneMinSecNumber  :=
    SectionNames[(latmin - cclatorg) mod 5,(cclonorg - lonmin - 1) mod 5];
  mn36 := latmin - cclatref;
  FtPerMinLat := LatFtPerMin[mn36];
  FtPerMinLon := LonFtPerMin[mn36];
  { do north or south line and footages }
  if fraclatmin < 0.50 then begin { return from south line }
    NSLineLetter   := 'S';
    FeetFromNSLine := round(FtPerMinLat * fraclatmin);
  end else begin { return from north line }
    NSLineLetter   := 'N';
    FeetFromNSLine := round(FtPerMinLat - (FtPerMinLat * fraclatmin));
  end;
  { do east or west line and footages }
  if fraclonmin < 0.50 then begin { return from east line }
    EWLineLetter   := 'E';
    FeetFromEWLine := round(FtPerMinLon * fraclonmin);
  end else begin { return from west line }
    EWLineLetter   := 'W';
    FeetFromEWLine := round(FtPerMinLon - (FtPerMinLon * fraclonmin));
  end;
 _LatLonToCC := true
end; { _LatLonToCC }

{ ************************************************************************
  function InKentucky

  Purpose: to check if a given Carter coordinate 5-minute section
      letter and number are in Kentucky

  Usage:

      boolean_variable := InKentucky(identifier_value,
                                     letter_value,
                                     number_value);

      identifier_value -- any string containing information useful
          for identifying which item is being processed

      letter_value -- a string value that is supposed to be a
          valid Carter coordinate 5-minute section letter

      number_value -- an integer value that is supposed to be a
          valid Carter coordinate 5-minute section number

  Errors detected:
      1) 5-minute section letter is not "A" to "Z" or "AA" to "GG"
      2) 5-minute section number is not from 0 to 91
      3) letter and number designate a 5-minute section that lies
         outside of Kentucky

  Returns: boolean value TRUE iff letter and number are valid and
      designated 5-minute section is in Kentucky

  Note:

      This routine is provided for the convenience of any users and
      to maintain a consistent implementation strategy.  It duplicates
      the function of routines provided elsewhere
}
function InKentucky;
var
  i : byte;
  ltrsym :  CarterLtrs;
begin
  InKentucky := true;
  { make letter upper case and strip out any non-alphabetic stuff }
  i:=length(SecLetter);
  while i>0 do begin
    SecLetter[i] := upcase(SecLetter[i]);
    if not (SecLetter[i] in ['A'..'Z']) then delete(SecLetter,i,1);
    i := i - 1;
  end;
  if not CvtToLtrSym(identifier,SecLetter,ltrsym) then begin
    InKentucky := false;
    exit;
  end;
  if (SecNumber < 0) or (SecNumber > 91) then begin
    writeln('Invalid 5-minute section number ',SecNumber,' at ',
      'id = ',identifier);
    InKentucky := false;
    exit;
  end;
  InKentucky := _InKentucky(identifier,ltrsym,SecNumber);
end;

{ ************************************************************************
  function CarterToLatLon

  Purpose: convert Carter coordinates to latitude and longitude

       Input consists of a string value assumed to be the Carter coordinate.
       This value is parsed and rigorously checked for errors.  The function
       will fail (return FALSE) if any errors are detected.

  Usage:

      boolean_variable := CarterToLatLon(identifier_string,
                                         carter_coordinate_string,
                                         latitude_variable,
                                         longitude_variable);

      identifier_string -- is any string containing information useful
              for identifying which item is having its latitude and
              longitude calculated should there be some error

      carter_coordinate_string -- is a string containing what is probably
              a Carter coordinate

      latitude_variable -- a real variable that will be assigned the latitude

      longitude_variable -- a real variable that will be assigned the
              longitude (Note: west longitudes, i.e., all those longitudes
              within Kentucky, are returned as negative)

  Errors detected:
      1) Invalid characters
      2) Missing parts of the Carter coordinate
      3) Footages from section lines not integer values
      4) Footages from section lines greater than maximum allowed (see
         FTPERMIN.PAS file)
      5) Footages not measured from adjacent pair of section lines
      6) Improper designation of what line a footage is measured from
      7) 1-minute section number <=0 or > 25
      8) Invalid 5-minute "quad" letter
      9) 5-minute quad number <0 or >91

  Returns: boolean value TRUE iff conversion succeeds

  The variables Latitude and Longitude are assigned useful numbers if
  _CCToLatLon succeeds, else they will be set to 0.0.

**************************************************************************
*                                                                        *
*     Any Carter coordinate that is otherwise valid but which lies       *
*     outside of Kentucky will not be detected as an error.  A latitude  *
*     and longitude will be calculated and this routine will return      *
*     with an indication of successful completion.                       *
*                                                                        *
**************************************************************************

}

function CarterToLatLon;

var
  token, s, temp : string;
  NSFeet, EWFeet : integer;
  NorS, EorW     : char;
  Sec            : Sections;
  Ltr            : CarterLtrs;
  Num            : Numbers;
  i, ErrCode     : integer;
  c              : char;
  succeeded      : boolean;

  { function Gettoken

    Purpose: to return the next part of the Carter coordinate string
        be it number, letter, or what

    Usage: local to CarterToLatLon, MAY NOT BE CALLED BY USER AS IMPLEMENTED

    Errors detected:
        1) invalid characters
        2) running out of parts of the Carter coordinate when some are
           expected

    Returns: boolean value TRUE iff there is a valid token

    Note: Gettoken always returns upper case letters }

  function Gettoken(var str, token : string; validset : charset) : boolean;
    begin { Gettoken }
      token := '';
      { BCN 8-Jul-1994:  Also ignore tab characters in input string }
      while str[1] in [' ',',','	'] do delete(str,1,1);
      if length(str)>0 then begin
        if not (upcase(str[1]) in validset) then begin
          writeln('Bad or unexpected character in Carter coordinate, "',str[1],
                  '" at id ',Identifier);
          Gettoken := false;
          exit
        end;
        repeat
          token := token + upcase(str[1]);
          delete(str,1,1)
        until (not (upcase(str[1]) in validset)) or (length(str)=0);
        Gettoken := true
      end else begin { we were looking for something and nothing was there }
        writeln('Carter coordinate not complete at id ',Identifier);
        Gettoken := false
      end;
      if debug then writeln('Current token = "',token,'"');
    end; { Gettoken }

  { function LineFootages

    Purpose: take the first two tokens from the input string "s" and convert
        them into a valid footage measurement and valid line designation

    Usage: local to CarterToLatLon, MAY NOT BE CALLED BY USER AS IMPLEMENTED

    Errors detected:
        1) can't make an integer footage out of the first token
        2) can't make a north, south, east, or west line designation out
           of the second token

    Returns: boolean value TRUE iff could do both of the above }

  function LineFootages(var s       : string;
                        var Footage : integer;
                        var LineLtr : char) : boolean;
    var
      token   : string;
      ErrCode : integer;
    begin { LineFootages }
      { get footage }
      if not Gettoken(s,token,['0'..'9']) then begin
        LineFootages := false;
        exit
      end;
      val(token,Footage,ErrCode);
      if ErrCode<>0 then begin
        writeln('Error in footage value "',token,'" at id ',Identifier);
        LineFootages := false;
        exit
      end;
      { get designation }
      if not Gettoken(s,token,['A'..'Z']) then begin
        LineFootages := false;
        exit
      end;
      LineLtr := ' ';
      if pos('N',token)>0 then LineLtr := 'N';
      if pos('S',token)>0 then LineLtr := 'S';
      if pos('E',token)>0 then LineLtr := 'E';
      if pos('W',token)>0 then LineLtr := 'W';
      if LineLtr=' ' then begin
        writeln('Error: can''t find N, S, E, or W in "',token,
                '" at id ',Identifier);
        LineFootages := false;
      end else LineFootages := true
    end; { LineFootages }

  { function DoFootages

    Purpose: convert the next four tokens in the input string to a
        pair of integer footage measurements from an adjacent pair of
        section lines

    Usage: local to CarterToLatLon, MAY NOT BE CALLED BY USER AS IMPLEMENTED

    Errors detected:
        1) adjacent section line footage missing for given first section line
           footage (see messages)

    Returns: boolean value TRUE iff a valid pair of adjacent section line
        footages is available }

  function DoFootages: boolean;
    begin
      DoFootages := true;
      { get footage and N,S,E, or W designation pair }
      if not LineFootages(s,i,c) then begin
        DoFootages := false;
        exit
      end;
      if c in ['N','S'] then begin
        NSFeet := i; NorS := c
      end else begin
        EWFeet := i; EorW := c
      end;

      { at this spot, some users may have an "x" meaning "by" here,
        kill any spaces or x's }
      while upcase(s[1]) in [' ','X'] do delete(s,1,1);

      { get next footage and designation pair and guarantee that this pair
        is adjacent to the previous pair }
      if not LineFootages(s,i,c) then begin
        DoFootages := false;
        exit
      end;
      if c in ['E','W'] then begin { is it east or west }
        if EorW=' ' then begin { better not have east/west already }
          EWFeet := i; EorW := c
        end else begin { we got east or west, but we've already got it }
          writeln('Error: north or south footages missing at id ',Identifier);
          DoFootages := false;
          exit
        end
      end else begin { no, we got north or south }
        if NorS=' ' then begin { better not have north/south already }
          NSFeet := i; NorS := c
        end else begin { we got north or south, but we've already got it }
          writeln('Error: east or west footages missing at id ',Identifier);
          DoFootages := false;
          exit
        end
      end;
    end; { DoFootages }

  { function DoCoordinates

    Purpose: convert the next three tokens in the input string to a
        Carter coordinate 1-minute section, 5-minute section letter,
        and 5-minute section number of the proper variable types

    Usage: local to CarterToLatLon, MAY NOT BE CALLED BY USER AS IMPLEMENTED

    Errors detected:
        1) section, letter, or number missing or invalid

    Returns: boolean value TRUE iff a section, letter, and number are
        valid and available }

  function DoCoordinates: boolean;
    begin
      DoCoordinates := true;
      { get the 1-minute section number }
      if not Gettoken(s,token,['0'..'9']) then begin
        writeln('Error: Can''t find 1-minute section number at id ',
                 Identifier);
        DoCoordinates := false;
        exit
      end;
      val(token,i,ErrCode);
      if ErrCode<>0 then begin
        writeln('Can''t convert "',token,'" to integer 1-minute section ',
                'number at id ',Identifier);
        DoCoordinates := false;
        exit
      end;
      if (i<1) or (i>25) then
      begin
        writeln('Error: 1-minute section number out of range at id ',
                Identifier);
        DoCoordinates := false;
        exit
      end else Sec := i;

      { there is the possibility of a dash here, kill it }
      while s[1] in [' ','-'] do delete(s,1,1);

      { get the 5-minute quad letter }
      if not Gettoken(s,token,['A'..'Z']) then begin
        writeln('Can''t find 5-minute section letter at id ',Identifier);
        DoCoordinates := false;
        exit
      end;
      if not CvtToLtrSym(identifier,token,Ltr) then begin
        writeln('Can''t convert "',token,'" to 5-minute section letter at id ',
                Identifier);
        DoCoordinates := false;
        exit
      end;

      { there is the possibility of a dash here, kill it }
      while s[1] in [' ','-'] do delete(s,1,1);

      { finally, get the 5-minute quad number }
      if not Gettoken(s,token,['0'..'9']) then begin
        writeln('Can''t find 5-minute section number at id ',Identifier);
        DoCoordinates := false;
        exit
      end;
      val(token,i,ErrCode);
      if ErrCode<>0 then begin
        writeln('Can''t convert "',token,'" to integer 5-minute section ',
                'number at id ',Identifier);
        DoCoordinates := false;
        exit
      end;
      if (i<0) or (i>91) then
      begin
        writeln('Error: 5-minute section number, ',i,
                ', out of range at id ',Identifier);
        DoCoordinates := false;
        exit
      end else Num := i;
    end; { DoCoordinates }

  { function FeetFirst

    Purpose: to determine whether footage measurements from section lines
        are first (default) or if the section, letter, and number are first
        in the input string

    Usage: local to CarterToLatLon, MAY NOT BE CALLED BY USER AS IMPLEMENTED

    Errors detected: none

    Returns: boolean value TRUE iff footages first can be determined }

  function FeetFirst(s: string; var IsFirst: boolptr): boolean;
    var
      temp, token : string;
      i,code      : integer;
      ok          : boolean;

      procedure Ignore(var s: string);
        begin
          { BCN 8-Jul-1994:  Also ignore tab character }
          while upcase(s[1]) in [' ','-','X','	'] do delete(s,1,1)
        end;

    begin
      new(IsFirst);
      IsFirst^  := true;
      FeetFirst := true;
      temp := s;
      ignore(temp);
      ok := Gettoken(temp,token,['0'..'9']); { 1st token is always numeric }
      if not Ok then begin
        FeetFirst := false;
        exit
      end;
      val(token,i,code);
      if code<>0 then begin
        FeetFirst := false;
        exit
      end;
      if (i=0) or (i>25) then exit; { first token was a footage }

      { if the next non-blank token is a dash, footages are not first }
      { BCN 14-Mar-1986 bug fix }
      i := 1;
      while temp[i]=' ' do i := i+1;
      if temp[i]='-' then begin
        IsFirst^ := false;
        exit
      end;

      { can't tell if first token was a small footage or a section number }
      ignore(temp);
      ok := Gettoken(temp,token,['A'..'Z']); { 2nd token will be alphabetic }
      if not ok then begin
        FeetFirst := false;
        exit
      end;
      if not (token[1] in ['E','F','N','S','W']) then begin
        IsFirst^ := false;
        exit
      end;

      { can't tell if second token was a line designation or a 5-minute
        quad letter }
      ignore(temp);
      ok := Gettoken(temp,token,['0'..'9']); { 3rd token will be numeric }
      if not Ok then begin
        FeetFirst := false;
        exit
      end;
      val(token,i,code);
      if code<>0 then begin
        FeetFirst := false;
        exit
      end;
      if i>91 then exit; { third token was a footage }

      { still can't tell, the fourth token is paydirt }
      ignore(temp);
      if temp[1] in ['A'..'Z'] then IsFirst^ := false
    end; { FeetFirst }

begin { CarterToLatLon }
  s         := CarterCoord;
  succeeded := true;
  NorS      := ' ';
  EorW      := ' ';
  Latitude  := 0.0;
  Longitude := 0.0;
  CarterToLatLon := false; { Bug fix 23-Jun-1989 BCN }

  if CarterFormKnown = nil then begin
    succeeded := FeetFirst(s,CarterFormKnown);
    if not succeeded then begin
      writeln('Error: can''t tell format of Carter coordinates');
      writeln('  Carter coordinate string "',s,'" invalid at id ',
        Identifier);
      writeln('  Program terminated.');
      halt
    end;
    if debug then
      if CarterFormKnown^
        then writeln('Format is feet first')
        else writeln('Format is section, letter, number first');
  end;

  if CarterFormKnown^ = true then begin
    succeeded := DoFootages;
    if not succeeded then exit;
    succeeded := DoCoordinates;
    if not succeeded then exit
  end else begin
    succeeded := DoCoordinates;
    if not succeeded then exit;
    succeeded := DoFootages;
    if not succeeded then exit
  end;

  { Now, we have all we need for a complete Carter coordinate location.
    Ignore the rest of the input string and try to convert what we have
    to latitude and longitude }
  if _CCToLatLon(Identifier,NSFeet,NorS,EWFeet,EorW,Sec,Ltr,Num,
                     Latitude, Longitude)
    then CarterToLatLon := true
    else CarterToLatLon := false;
end; { CarterToLatLon }

{ *********************************************************************
  function LatLonToCarter

  Purpose: convert latitude/longitude location to Carter coordinate

      Input consists of a latitude and longitude in decimal degrees
      and the output is one of two user selected formats of
      Carter coordinate strings.

  Usage:

      boolean_variable := LatLonToCarter(identifier_string,
                                         carter_coordinate_variable,
                                         latitude_value,
                                         longitude_value,
                                         boolean_format_selector_value);

      identifier_string -- is a string containing information useful
          for identifying which item is being processed should there
          be an error

      carter_coordinate_variable -- the output Carter coordinate

      latitude_value -- the latitude of the location in decimal degrees

      longitude_value -- the longitude of the location in decimal degrees

      boolean_format_selector_value -- if true, an expanded version of
          the Carter coordinate will be returned, i.e.,

              1234FNL x 2345FEL 12- A-25

          else (value is false) a compressed format like that used at
          KGS to store the Carter coordinate will be returned, i.e.,

              1234FNL2345FEL12 A25

          A global typed constant, expand, is defined and initially set to
          false is defined in this unit

  Errors detected:

      A message is issued if the location is not in Kentucky

  Returns: boolean value TRUE iff conversion succeeds

      The output Carter coordinate string is set to the null string
      on failure of the routine
}
function LatLonToCarter;
var
  fns, few : integer;
  ns, ew   : char;
  sec      : sections;
  ltr      : carterltrs;
  num      : numbers;
  temp     : string;

  function lpad(s : string; l : integer; ch : string) : string;
  { pad input string s to length l with character ch }
  begin
    while length(s) < l do s := ch + s;
    lpad := s;
  end;

begin
  CarterCoord := '';
  if _LatLonToCC(Identifier,fns,ns,few,ew,sec,
                 ltr,num,Latitude,Longitude) then begin
    if not expand then begin
      str(fns,temp); temp := lpad(temp,4,'0');
      CarterCoord := temp + 'F' + upcase(ns) + 'L';
      str(few,temp); temp := lpad(temp,4,'0');
      CarterCoord := CarterCoord + temp + 'F' + upcase(ew) + 'L';
      str(sec,temp); temp := lpad(temp,2,'0');
      CarterCoord := CarterCoord + temp + lpad(CarterRep[ltr],2,' ');
      str(num,temp); temp := lpad(temp,2,'0');
      CarterCoord := CarterCoord + temp;
    end else begin
      str(fns,temp); temp := lpad(temp,4,'0');
      CarterCoord := temp + ' F' + upcase(ns) + 'L x ';
      str(few,temp); temp := lpad(temp,4,'0');
      CarterCoord := CarterCoord + temp + ' F' + upcase(ew) + 'L ';
      str(sec,temp); temp := lpad(temp,2,'0');
      CarterCoord := CarterCoord + temp + '-' + lpad(CarterRep[ltr],2,' ');
      str(num,temp); temp := lpad(temp,2,'0');
      CarterCoord := CarterCoord + '-' + temp;
    end;
    LatLonToCarter := true;
  end else LatLonToCarter := false;
end;
begin
  CarterFormKnown := nil;  { initialize to don't know form }
end. { carterll unit }

